#define NRF_LOG_MODULE_NAME main
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
NRF_LOG_MODULE_REGISTER();

#include <stdint.h>
#include <assert.h>
#include "boards.h"
#include "nrf_mbr.h"
#include "nrf_dfu_flash.h"
#include "nrf_bootloader_app_start.h"
#include "app_error.h"
#include "nrf_bootloader_info.h"
#include "nrf_delay.h"
#include "nrf_log_default_backends.h"
#include "spi_flash_block.h"
#include "crc32.h"
#include "nrf_dfu_utils.h"




// The address offset from the beginning of (external) Flash where the DFU header is stored.
static const uint32_t WA_DFU_HEADER_OFFSET = 0;

struct wa_dfu_header
{
    size_t   size;
    size_t   offset;
    uint32_t crc;
    uint32_t good;
};



static void on_error(void)
{
    NRF_LOG_FINAL_FLUSH();

#if NRF_MODULE_ENABLED(NRF_LOG_BACKEND_RTT)
    // To allow the buffer to be flushed by the host.
    nrf_delay_ms(100);
#endif
#ifdef NRF_DFU_DEBUG_VERSION
    NRF_BREAKPOINT_COND;
#endif
    NVIC_SystemReset();
}

void app_error_handler(uint32_t error_code, uint32_t line_num, const uint8_t * p_file_name)
{
    NRF_LOG_ERROR("%s:%d", p_file_name, line_num);
    on_error();
}

void app_error_fault_handler(uint32_t id, uint32_t pc, uint32_t info)
{
    NRF_LOG_ERROR("Received a fault! id: 0x%08x, pc: 0x%08x, info: 0x%08x", id, pc, info);
    on_error();
}

void app_error_handler_bare(uint32_t error_code)
{
    NRF_LOG_ERROR("Received an error: 0x%08x!", error_code);
    on_error();
}




static void log_header(const struct wa_dfu_header * header)
{
    NRF_LOG_DEBUG("[DFU Header] size: %u, offset: 0x%X, crc: 0x%08X, good: %u", header->size, header->offset, header->crc, header->good);
}




static struct wa_dfu_header get_internal_flash_dfu_header()
{
    struct wa_dfu_header header = {};
    memcpy(&header, (const void *)BOOTLOADER_SETTINGS_ADDRESS, sizeof(header));
    return header;
}

static struct wa_dfu_header get_spi_flash_dfu_header()
{
    struct wa_dfu_header header = {};
    const ret_code_t error = spi_flash_block_read_data(WA_DFU_HEADER_OFFSET, &header, sizeof(header));
    if (error)
    {
        NRF_LOG_ERROR("Error %u: Failed to read DFU header from SPI Flash!", error);
    }
    return header;
}

static bool headers_are_equal(const struct wa_dfu_header * header1, const struct wa_dfu_header * header2)
{
    return memcmp(header1, header2, sizeof(struct wa_dfu_header)) == 0;
}

static bool dfu_is_needed(const struct wa_dfu_header * internal_header, const struct wa_dfu_header * spi_header)
{
    return ! headers_are_equal(internal_header, spi_header);
}

static ret_code_t check_dfu_size(const struct wa_dfu_header * dfu_header)
{
    NRF_LOG_INFO("Checking DFU size...");

    const size_t max_size = BOOTLOADER_START_ADDR - nrf_dfu_app_start_address();
    if (dfu_header->size > max_size)
    {
        NRF_LOG_ERROR("Error: DFU firmware image is too large! DFU will not be performed. DFU firmware image size: %u, max size: %u", dfu_header->size, max_size);
        return NRF_ERROR_INVALID_LENGTH;
    }

    NRF_LOG_INFO("Checking DFU size... done");
    return NRF_SUCCESS;
}

static ret_code_t erase_firmware_header(void)
{
    NRF_LOG_INFO("Erasing firmware header...");

    ret_code_t error = nrf_dfu_flash_erase(BOOTLOADER_SETTINGS_ADDRESS, ALIGN_TO_PAGE(BOOTLOADER_SETTINGS_PAGE_SIZE) / CODE_PAGE_SIZE, /*callback*/NULL);
    if (error)
    {
        NRF_LOG_ERROR("Error %u: Failed erasing firmware header!", error);
        return error;
    }

    NRF_LOG_INFO("Erasing firmware header... done");
    return NRF_SUCCESS;
}

static ret_code_t erase_firmware_image(const struct wa_dfu_header * dfu_header)
{
    NRF_LOG_INFO("Erasing firmware image...");

    ret_code_t error = nrf_dfu_flash_erase(nrf_dfu_app_start_address(), ALIGN_TO_PAGE(dfu_header->size) / CODE_PAGE_SIZE, /*callback*/NULL);
    if (error)
    {
        NRF_LOG_ERROR("Error %u: Failed erasing firmware image!", error);
        return error;
    }

    NRF_LOG_INFO("Erasing firmware header... done");
    return NRF_SUCCESS;
}

static ret_code_t copy_firmware_image(const struct wa_dfu_header * dfu_header)
{
    NRF_LOG_INFO("Copying firmware image...");

    uint8_t buf[CODE_PAGE_SIZE];
    size_t index = 0;

    while (index < dfu_header->size)
    {
        const size_t remaining_size = dfu_header->size - index;
        const size_t copy_size = MIN(sizeof(buf), remaining_size);
        const uint32_t source_address = dfu_header->offset + index;
        const uint32_t destination_address = nrf_dfu_app_start_address() + index;
        NRF_LOG_DEBUG("index: %u, remaining size: %u, copy size: %u, source addr: 0x%08X, dest addr: 0x%08X", index, remaining_size, copy_size, source_address, destination_address);
        if (NRF_LOG_DEFAULT_LEVEL >= 4) NRF_LOG_FLUSH();


        ret_code_t error = spi_flash_block_read_data(source_address, buf, copy_size);
        if (error)
        {
            NRF_LOG_ERROR("Error %u: DFU failed while reading firmware image from SPI Flash! Address: 0x%08X, size: %u", source_address, copy_size);
            return error;
        }

        error = nrf_dfu_flash_store(destination_address, buf, copy_size, /*callback*/NULL);
        if (error)
        {
            NRF_LOG_ERROR("Error %u: DFU failed while writing firmware image to Flash! Address: 0x%08X, size: %u", destination_address, copy_size);
            return error;
        }

        index += copy_size;
    }

    NRF_LOG_INFO("Copying firmware image... done");
    return NRF_SUCCESS;
}

static ret_code_t check_firmware_image(const struct wa_dfu_header * dfu_header)
{
    NRF_LOG_INFO("Checking firmware image...");

    const uint32_t crc = crc32_compute((const uint8_t *)nrf_dfu_app_start_address(), dfu_header->size, NULL);
    if (crc != dfu_header->crc)
    {
        NRF_LOG_ERROR("Error: Firmware image CRCs do not match! Firmware image CRC: 0x%X, DFU image CRC: 0x%X", crc, dfu_header->crc);
        return NRF_ERROR_INVALID_DATA;
    }

    NRF_LOG_INFO("Checking firmware image... done");
    return NRF_SUCCESS;
}

static ret_code_t copy_firmware_header(const struct wa_dfu_header * dfu_header)
{
    NRF_LOG_INFO("Copying firmware header...");

    ret_code_t error = nrf_dfu_flash_store(BOOTLOADER_SETTINGS_ADDRESS, dfu_header, sizeof(struct wa_dfu_header), /*callback*/NULL);
    if (error)
    {
        NRF_LOG_ERROR("Error %u: Failed erasing firmware image!", error);
        return error;
    }

    NRF_LOG_INFO("Copying firmware header... done");
    return NRF_SUCCESS;
}

static ret_code_t update_device_firmware(const struct wa_dfu_header * spi_header)
{
    NRF_LOG_INFO("Updating firmware...");

    VERIFY_SUCCESS(check_dfu_size(spi_header));
    VERIFY_SUCCESS(erase_firmware_header());
    VERIFY_SUCCESS(erase_firmware_image(spi_header));
    VERIFY_SUCCESS(copy_firmware_image(spi_header));
    VERIFY_SUCCESS(check_firmware_image(spi_header));
    VERIFY_SUCCESS(copy_firmware_header(spi_header));
    
    NRF_LOG_INFO("Updating firmware... done");
    return NRF_SUCCESS;
}

static void print_region_info(const char * region_name, uint32_t start_address, size_t size)
{
    const uint32_t end_address = start_address + size;
    NRF_LOG_DEBUG("%20s: address [0x%08X, 0x%08X), size: 0x%08X (%u)", region_name, start_address, end_address, size, size);
    NRF_LOG_FLUSH();
}


int main(void)
{
    ret_code_t ret_val;

    // Protect MBR and bootloader code from being overwritten.
    ret_val = nrf_bootloader_flash_protect(0, MBR_SIZE, /*read protect*/false);
    APP_ERROR_CHECK(ret_val);
    ret_val = nrf_bootloader_flash_protect(BOOTLOADER_START_ADDR, BOOTLOADER_SIZE, /*read protect*/false);
    APP_ERROR_CHECK(ret_val);

    (void) NRF_LOG_INIT(NULL);
    NRF_LOG_DEFAULT_BACKENDS_INIT();

    NRF_LOG_INFO("Running bootloader...");
    print_region_info("MBR", 0, MBR_SIZE);
    print_region_info("SoftDevice", MBR_SIZE, (nrf_dfu_app_start_address() - MBR_SIZE));
    print_region_info("Application", nrf_dfu_app_start_address(), (BOOTLOADER_START_ADDR - nrf_dfu_app_start_address()));
    print_region_info("Bootloader", BOOTLOADER_START_ADDR, BOOTLOADER_SIZE);
    print_region_info("MBR Settings", NRF_MBR_PARAMS_PAGE_ADDRESS, CODE_PAGE_SIZE);
    print_region_info("Bootloader Settings", BOOTLOADER_SETTINGS_ADDRESS, BOOTLOADER_SETTINGS_PAGE_SIZE);
    

    ret_val = nrf_dfu_flash_init(/*SD IRQ initialized*/false);
    APP_ERROR_CHECK(ret_val);

    spi_flash_block_init();

    const struct wa_dfu_header internal_header = get_internal_flash_dfu_header();
    const struct wa_dfu_header spi_header = get_spi_flash_dfu_header();
    log_header(&internal_header);
    log_header(&spi_header);
    
    if (dfu_is_needed(&internal_header, &spi_header))
    {
        ret_val = update_device_firmware(&spi_header);
        APP_ERROR_CHECK(ret_val);
    }

    spi_flash_block_uninit();

    NRF_LOG_INFO("Starting app...");
    NRF_LOG_FLUSH();
    nrf_bootloader_app_start();

    // Should never be reached.
    NRF_LOG_INFO("Running bootloader... done");
    return 0;
}
